"""Time related constants."""

__all__ = ['SEC_PER_DAY', 'SEC_PER_HOUR', 'SEC_PER_MIN', 'MONTH_NAMES']

SEC_PER_DAY = 86_400
SEC_PER_HOUR = 3_600
SEC_PER_MIN = 60
MONTH_NAMES = (
    None,
    'January', 'February', 'March', 'April', 'May', 'June',
    'July', 'August', 'September', 'October', 'November', 'December')
